import java.io.*;
import java.util.*;
import ParseForest;


/******************************************/
/*                                        */
/* Class: Chart                           */
/*                                        */
/******************************************/

public class Chart {
  public Field table[][]; 

  private Vector parseForest;
  private int currentIndex;  /* index of the current parse forest */
  private int currentSize;   /* number of parse forests read so far */
  private boolean readingFinished;
  private PushbackInputStream file;



  /******************************************/
  /*                                        */
  /* Constructor: Chart                     */
  /*                                        */
  /******************************************/

  public Chart( String[] args ) throws FileNotFoundException, IOException {

    readingFinished = false;
    currentSize = 0;
    parseForest = new Vector();

    if (args.length == 0)
      file = new PushbackInputStream(System.in, 100);
    else
      file = new PushbackInputStream( new BufferedInputStream( 
               new FileInputStream (args[0])), 100);

    select(0);
  }

  /******************************************/
  /*                                        */
  /* Access Methods                         */
  /*                                        */
  /******************************************/

  public ParseForest selectedParseForest() 
  {
    return (ParseForest)parseForest.get(currentIndex);
  }

  public boolean withHeads() 
  {
    return selectedParseForest().withHeads;
  }

  public boolean withScores() 
  {
    return selectedParseForest().withScores;
  }

  public int sentLength()
  {
    return selectedParseForest().sentLength();
  }

  public String word(int i)
  {
    return selectedParseForest().word[i];
  }

  public int currentIndex()
  {
    return currentIndex;
  }

  public int currentSize()
  {
    return currentSize;
  }


  /******************************************/
  /*                                        */
  /* Subclass: Field                        */
  /*                                        */
  /******************************************/

  public class Field {
    public ParseForest.Node[] node;
    public ParseForest.Node selectedNode;

    public Field( LinkedList l )
    {
      node = new ParseForest.Node[l.size()];
      ListIterator iter = l.listIterator(0);
      for( int i=0; iter.hasNext(); i++ )
	node[i] = (ParseForest.Node) iter.next();

      if (node.length > 0) {
	Arrays.sort(node, node[0]);
	selectedNode = node[0];
      }
    }
  }

  /******************************************/
  /*                                        */
  /* Method: select                         */
  /*                                        */
  /******************************************/

  void select( int index ) throws IOException, IndexOutOfBoundsException
  {

    /* read up to the successor of the requested parse forest */
    for( ; !readingFinished && currentSize<=index+1; currentSize++ ) {
      System.err.println("reading parse "+(currentSize+1));
      ParseForest p = new ParseForest(file);
      if (!p.unparsed && p.sentLength() == 0) {
	readingFinished = true;
	System.err.println("all "+currentSize+" parses read");
	break;
      }
      parseForest.add(p);
    }
    if (index < 0 || index >= currentSize)
      throw new IndexOutOfBoundsException();

    currentIndex = index;
    ParseForest p = selectedParseForest();
    if (!p.unparsed) {
      LinkedList[][] t = new LinkedList[p.sentLength()][p.sentLength()];
      for( int i=0; i<p.sentLength(); i++ )
	for( int k=0; k<p.sentLength()-i; k++ ) {
	  t[i][k] = new LinkedList();
	}

      for( int i=0; i<p.node.length; i++ ) {
	ParseForest.Node node = p.node[i];
	t[node.start][node.end-node.start].add (node);
      }
      
      /* convert linked lists to arrays */
      
      table = new Field[p.sentLength()][p.sentLength()];
      for( int i=0; i<p.sentLength(); i++ )
	for( int k=0; k<p.sentLength()-i; k++ )
	  table[i][k] = new Field(t[i][k]);
    }
  }
}
