import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import Chart;


/******************************************/
/*                                        */
/* Class: ChartDisplay                    */
/*                                        */
/******************************************/

public class ChartDisplay extends JPanel 
{

  private Chart chart;
  private JLabel wordLabel[];
  private CatButton catButton[][];
  private JPopupMenu popup;
  private ButtonEventListener buttonEL;
  private Button0EventListener button0EL;
  private MenuEventListener menuEL;
  private SubmenuEventListener submenuEL;


  /******************************************/
  /*                                        */
  /* Constructor: ChartDisplay              */
  /*                                        */
  /******************************************/

  public ChartDisplay( Chart ch )
  {
    super(new FlowLayout(FlowLayout.LEFT));
    buttonEL = new ButtonEventListener();
    button0EL = new Button0EventListener();
    menuEL = new MenuEventListener();
    submenuEL = new SubmenuEventListener();
    chart = ch;
    wordLabel = new JLabel[chart.sentLength()];
    catButton = new CatButton[chart.sentLength()][chart.sentLength()];

    /* Create the buttons */
    for( int i=0; i<chart.sentLength(); i++) {
      wordLabel[i] = new JLabel(" "+chart.word(i)+" ");
      wordLabel[i].setOpaque(true);
      for( int k=0; k<chart.sentLength()-i; k++) {
	catButton[i][k] = new CatButton();
	catButton[i][k].setMargin(new Insets(0,0,0,0));
	if (chart.table[i][k].node.length > 0) {
	  ParseForest.Node node = chart.table[i][k].selectedNode;
	  catButton[i][k].modify(node, false);
	  catButton[i][k].addMouseListener(buttonEL);
	  catButton[i][k].setToolTipText("Click to select analysis");
	}
	else {
	  catButton[i][k].setToolTipText("Click to dehighlight nodes");
	  catButton[i][k].addMouseListener(button0EL);
	  catButton[i][k].addActionListener(new AbstractAction () {
	    public void actionPerformed( ActionEvent e ) {
	      dehighlightAll();
	    }
	  });
	}
      }
    }

    /* Create the panel and add the buttons */

    GridBagLayout gridbag = new GridBagLayout();
    GridBagConstraints c = new GridBagConstraints();

    JPanel panel = new JPanel(gridbag);
    panel.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
    add(panel);
	
    c.fill = GridBagConstraints.BOTH;
    for( int i=0; i<chart.sentLength(); i++) {
      c.gridy = i;
      c.gridx = 0;
      gridbag.setConstraints(wordLabel[i], c);
      panel.add(wordLabel[i]);

      c.ipadx = 0;
      for( int k=0; k<chart.sentLength()-i; k++) {
	c.gridx++;
	gridbag.setConstraints(catButton[i][k], c);
	panel.add(catButton[i][k]);
      }
    }
  }

  /******************************************/
  /*                                        */
  /* Subclass: CatButton                    */
  /*                                        */
  /******************************************/

  private class CatButton extends JButton {

    boolean highlit;
    Border  defaultBorder;
    Border  highlitBorder;

    /******************************************/
    /*                                        */
    /* Constructor: CatButton                 */
    /*                                        */
    /******************************************/

    public CatButton() {
      super();
//       defaultBorder = (new JButton()).getBorder();
      defaultBorder = BorderFactory.createEtchedBorder(Color.yellow,Color.yellow);
      highlitBorder = BorderFactory.createEtchedBorder(Color.red,Color.red);
    }

    /******************************************/
    /*                                        */
    /* Method: setHighlight                   */
    /*                                        */
    /******************************************/

    public void setHighlight( boolean h ) {
      if (h != highlit) {
	if (h)
	  setBorder(highlitBorder);
	else
	  setBorder(defaultBorder);
	highlit = h;
      }
    }

    /******************************************/
    /*                                        */
    /* Method: modify                         */
    /*                                        */
    /******************************************/

    private void modify( ParseForest.Node node, boolean highlight )
    {
      setText(pad(node.cat));
      if (chart.selectedParseForest().withScores) {
	double f = Math.sqrt(node.score);
	setBackground(new Color(0.0f, (float)Math.sqrt(f), (float)Math.sqrt(1.0-f)));
      }
      setHighlight( highlight );
    }
  }


  /******************************************/
  /*                                        */
  /* Method: dehighlightAll                 */
  /*                                        */
  /******************************************/

  private void dehighlightAll() {
    for( int i=0; i<chart.sentLength(); i++)
      for( int k=0; k<chart.sentLength()-i; k++)
	catButton[i][k].setHighlight(false);
  }


  /******************************************/
  /*                                        */
  /* Method: pad                            */
  /*                                        */
  /******************************************/

  private String pad( String s )
  {
    switch (s.length()) {
    case 1: return "    "+s+"    ";
    case 2: return "   "+s+"   ";
    case 3: return "  "+s+"  ";
    case 4: return " "+s+" ";
    default: return s;
    }
  }

  /******************************************/
  /*                                        */
  /* Subclass: ButtonPosition               */
  /*                                        */
  /******************************************/

  private class ButtonPosition {
    public int x;
    public int y;

    public ButtonPosition( int y, int x ) {
      this.x = x;
      this.y = y;
    }
  }

  /******************************************/
  /*                                        */
  /* Method: getPosition                    */
  /*                                        */
  /******************************************/

  private ButtonPosition getPosition(Object source) {
    for( int i=0; i<chart.sentLength(); i++)
      for( int k=0; k<chart.sentLength()-i; k++)
	if (source == catButton[i][k]) {
	  return new ButtonPosition(i,k);
	}
    return null;
  }

  /******************************************/
  /*                                        */
  /* Subclass: Button0EventListener         */
  /*                                        */
  /******************************************/

  private class Button0EventListener extends MouseAdapter {

    public void mouseEntered(MouseEvent e) {
      Object source=e.getSource();
      ButtonPosition pos = getPosition(source);
      if (pos != null)
	for( int i=0; i<chart.sentLength(); i++) {
	  if (i < pos.y || i > pos.x+pos.y)
	    wordLabel[i].setBackground(getBackground());
	  else
	    wordLabel[i].setBackground(new Color(240,240,240));
	}
    }
  }

  /******************************************/
  /*                                        */
  /* Subclass: ButtonEventListener          */
  /*                                        */
  /******************************************/

  private class ButtonEventListener extends Button0EventListener {

    public void mousePressed(MouseEvent e) {
      Object source=e.getSource();
      ButtonPosition pos = getPosition(source);
      showPopup(e.getComponent(), e.getX(), e.getY(), pos);
    }

    public void mouseExited(MouseEvent e) {
      for( int i=0; i<chart.sentLength(); i++)
	wordLabel[i].setBackground(getBackground());
    }
  }

  /******************************************/
  /*                                        */
  /* Subclass: MenuEventListener            */
  /*                                        */
  /******************************************/

  private class MenuEventListener extends AbstractAction {

    public void actionPerformed(ActionEvent e) {
      StringTokenizer st = new StringTokenizer(e.getActionCommand());
      int y = Integer.parseInt(st.nextToken());
      int x = Integer.parseInt(st.nextToken());
      int i = Integer.parseInt(st.nextToken());
      dehighlightAll();
      selectNode(y,x,i);
    }
  }


  /******************************************/
  /*                                        */
  /* Subclass: SubmenuEventListener         */
  /*                                        */
  /******************************************/

  private class SubmenuEventListener extends AbstractAction {

    public void actionPerformed(ActionEvent e) {
      StringTokenizer st = new StringTokenizer(e.getActionCommand());
      int y = Integer.parseInt(st.nextToken());
      int x = Integer.parseInt(st.nextToken());
      int i = Integer.parseInt(st.nextToken());
      int k = Integer.parseInt(st.nextToken());
      dehighlightAll();
      selectAnalysis(y,x,i,k);
    }
  }


  /******************************************/
  /*                                        */
  /* Method: selectNode                     */
  /*                                        */
  /******************************************/

  private ParseForest.Node selectNode( int y, int x, int n )
  {
    Chart.Field f = chart.table[y][x];
    f.selectedNode = f.node[n];
    catButton[y][x].modify(f.selectedNode, true);
    return f.selectedNode;
  }


  /******************************************/
  /*                                        */
  /* Method: selectAnalysis                 */
  /*                                        */
  /******************************************/

  private void selectAnalysis( int y, int x, int n, int m )
  {
    ParseForest.Node node = selectNode( y, x, n );
    ParseForest.Analysis ana = node.analysis[m];
    for( int i=0; i<ana.daughter.length; i++) {
      node = chart.selectedParseForest().node[ana.daughter[i]];
      y = node.start;
      x = node.end - node.start;
      Chart.Field f = chart.table[y][x];
      f.selectedNode = node;
      catButton[y][x].modify(f.selectedNode, true);
    }
  }

  /******************************************/
  /*                                        */
  /* Method: showPopup                      */
  /*                                        */
  /******************************************/

  private void showPopup( Component c, int x, int y, ButtonPosition pos){

    popup = new JPopupMenu();

    for( int i=0; i<chart.table[pos.y][pos.x].node.length; i++) {
      ParseForest.Node node = chart.table[pos.y][pos.x].node[i];
      String s = node.cat;
      if (chart.withScores())
	s = ""+format(node.score)+" "+s;
      if (chart.withHeads())
	s += " "+node.head;

      if (node.terminal) {
	JMenuItem m = new JMenuItem(s);
	  m.setActionCommand(Integer.toString(pos.y) + " " +
			     Integer.toString(pos.x) + " " + 
			     Integer.toString(i));
	m.addActionListener(menuEL);
	popup.add(m);
      }
      else {
	JMenu menu = new JMenu(s);
	for( int k=0; k<node.analysis.length; k++) {
	  ParseForest.Analysis ana = node.analysis[k];
	  
	  if (chart.withScores())
	    s = format(ana.score) + " ";
	  else
	    s = "";
	  for( int j=0; j<ana.daughter.length; j++)
	    s += chart.selectedParseForest().node[ana.daughter[j]].cat + " ";
	  s += " "+(node.start+1);
	  for( int j=0; j<ana.daughter.length; j++)
	    s += "-"+(chart.selectedParseForest().node[ana.daughter[j]].end+1);
	  JMenuItem m = new JMenuItem(s);
	  m.setActionCommand(Integer.toString(pos.y) + " " +
			     Integer.toString(pos.x) + " " + 
			     Integer.toString(i) + " " +
			     Integer.toString(k));
	  m.addActionListener(submenuEL);
	  menu.add(m);
	}
	popup.add(menu);
      }
    }

    popup.show( c, x, y );
  }


  /******************************************/
  /*                                        */
  /* Method: format                         */
  /*                                        */
  /******************************************/

  private String format( double f ) {
    int n = (int)(f * 10000 + 0.5);
    if (n >= 10000)
      return "1.0000";
    if (n >= 1000)
      return "0." + n;
    else if (n >= 100)
      return "0.0" + n;
    else if (n >= 10)
      return "0.00" + n;
    else
      return "0.000" + n;
  }
}

