import java.io.*;
import java.util.*;
import VarArray;

/******************************************/
/*                                        */
/* Class: ParseForest                     */
/*                                        */
/******************************************/

public class ParseForest {

  public boolean withHeads, withScores, unparsed;
  public Node node[];
  public String word[];

  private boolean withHeadsDefined, withScoresDefined;
  private VarArray wordBuffer;
  private static int NONE = -1;
  private static double NOSCORE = -1.0;

  /******************************************/
  /*                                        */
  /* Constructor: ParseForest               */
  /*                                        */
  /******************************************/

  public ParseForest( PushbackInputStream file )  throws IOException {
    LinkedList buffer = new LinkedList();
    ListIterator iter;

    /* check whether the next sentence was parsed successfully */
    String s = readString(file);

    if (s.equals("parsing")) {
      String s2 = readString(file);
      if (s2.equals("failed!")) {
	unparsed = true;
	return;
      }
      else {
	pushbackToken(s2, file);
      }
    }
    pushbackToken(s, file);
    
    wordBuffer = new VarArray();
    for(;;) {
      Node n = new Node(file);
      if (n.cat.length() == 0)
	break;
      buffer.add(n);
      if (n.lastNode)
	break;
    }
    
    /* convert linked lists to arrays */
    
    word = new String[wordBuffer.size];
    for( int i=0; i<word.length; i++ ) {
      word[i] = (String) wordBuffer.get(i);
    }

    node = new Node[buffer.size()];
    iter = buffer.listIterator(0);
    for( int i=0; iter.hasNext(); i++ )
      node[i] = (Node) iter.next();
  } 


  /******************************************/
  /*                                        */
  /* Method: pushbackToken                  */
  /*                                        */
  /******************************************/

  private void pushbackToken( String s, PushbackInputStream file )
       throws IOException
  {
    file.unread(' ');
    for( int i=s.length()-1; i>=0; i--) {
      file.unread(s.charAt(i));
    }
  }

  /******************************************/
  /*                                        */
  /* Method: unquote                        */
  /*                                        */
  /******************************************/

  public int sentLength()
  {
    return word.length;
  }
    


  /******************************************/
  /*                                        */
  /* Subclass: Node                         */
  /*                                        */
  /******************************************/

  public class Node implements Comparator {

    boolean terminal;
    boolean lastNode;
    public int start, end;
    public String cat;
    public double score;
    public String head;
    public Analysis[] analysis;

    /******************************************/
    /*                                        */
    /* Constructor: Node                      */
    /*                                        */
    /******************************************/

    public Node( PushbackInputStream file ) throws IOException {
      cat = readString(file);
      score = NOSCORE;
//       System.err.println("cat: "+cat);
      if (cat.length() != 0) {
	String s;

	start = Integer.parseInt(readString(file));
	end = Integer.parseInt(readString(file))-1;
// 	System.err.println("span: "+start+" "+end);
	s = readString(file);
	if (s.length() != 0) {
	  head = unquote(s);
	  if (head != null) {
	    s = readString(file);
// 	    System.err.println("head: "+head);
	  }
	}
	if (s.length() != 0) {
	  score = Double.parseDouble(s);
// 	  System.err.println("score: "+score);
	  s = readString( file );
	  if (s.length() != 0)
	    throw new IOException("Error in input at: " + s);
	}

	/* check and set flags */

	if (withHeadsDefined) {
	  if ((head != null && !withHeads) || (head == null && withHeads))
	    throw new IOException("Error in input: missing head");
	}
	else {
	  withHeadsDefined = true;
	  withHeads = (head != null);
	}
	
	if (withScoresDefined) {
	  if ((score != NOSCORE && !withScores) ||
	      (score == NOSCORE && withScores))
	    throw new IOException("Error in input: missing score");
	}
	else {
	  withScoresDefined = true;
	  withScores = (score != NOSCORE);
	}
	
	/* check whether this is a terminal node */

	s = readTerm( file );
	if (s.length() > 0) {
	  terminal = true;
	  wordBuffer.set(s, start);
// 	  System.err.println("word: "+s);
	  s = readString(file);
	}
	else {
	  /* nonterminal node */
	  LinkedList buffer = new LinkedList();

	  for(;;) {
	    Analysis a = new Analysis(file);
	    buffer.add(a);
	    s = readString(file);
	    if (!s.equals("%"))
	      break;
	  }

	  /* convert linked list to array */

	  analysis = new Analysis[buffer.size()];
	  ListIterator iter = buffer.listIterator(0);
	  for( int i=0; iter.hasNext(); i++ )
	    analysis[i] = (Analysis) iter.next();

	  if (analysis.length > 1)
	    Arrays.sort(analysis, analysis[0]);
	}

	if (s.equals("%%%"))
	  lastNode = true;
	else if (s.equals("%%"))
	  ;
	else if (!s.equals("%"))
	  throw new IOException("Error in input at "+s);
      }
    }

    /******************************************/
    /*                                        */
    /* Method: equals                         */
    /*                                        */
    /******************************************/

    public boolean equals( Node n )
    {
      return (n.score == score && n.cat.equals(cat));
    }

    /******************************************/
    /*                                        */
    /* Method: compare                        */
    /*                                        */
    /******************************************/

    public int compare( Object o1, Object o2 )
    {
      Node n1 = (Node)o1;
      Node n2 = (Node)o2;
      if (n1.score > n2.score)
	return -1;
      if (n1.score < n2.score)
	return 1;
      return (n1.cat.compareTo(n2.cat));
    }
  }

  /******************************************/
  /*                                        */
  /* Subclass: Analysis                     */
  /*                                        */
  /******************************************/

  public class Analysis implements Comparator {

    private int ruleNumber;
    public double score;
    public int[] daughter;

    /******************************************/
    /*                                        */
    /* Constructor: Analysis                  */
    /*                                        */
    /******************************************/

    public Analysis( PushbackInputStream file ) throws IOException {
      String s = readString(file);
      LinkedList buffer = new LinkedList();

      ruleNumber = Integer.parseInt(s);
//       System.err.println("rule: "+ruleNumber);
      if (withScores) {
	score = Double.parseDouble(readString(file));
// 	System.err.println("score: "+score);
      }
      else
	score = NOSCORE;

      for( int i=0; ; i++) {
	s = readString(file);
	if (s.charAt(0) == '%')
	  break;
// 	System.err.println("daughter["+i+"]: "+s);
	buffer.add(new Integer(s));
      }

      pushbackToken(s, file);

      /* convert linked list to array */

      daughter = new int[buffer.size()];
      ListIterator iter = buffer.listIterator(0);
      for( int i=0; iter.hasNext(); i++ )
	daughter[i] = ((Integer) iter.next()).intValue();
    }

    /******************************************/
    /*                                        */
    /* Method: equals                         */
    /*                                        */
    /******************************************/

    public boolean equals( Analysis a )
    {
      return (a.ruleNumber != ruleNumber);
    }

    /******************************************/
    /*                                        */
    /* Method: compare                        */
    /*                                        */
    /******************************************/

    public int compare( Object o1, Object o2 )
    {
      Analysis a1 = (Analysis)o1;
      Analysis a2 = (Analysis)o2;

      if (withScores) {
	if (a1.score > a2.score)
	  return -1;
	else if (a1.score < a2.score)
	  return 1;
      }
      for( int i=0; i<a1.daughter.length && i<a2.daughter.length; i++)
	if (a1.daughter[i] > a2.daughter[i])
	  return -1;
	else if (a1.daughter[i] < a2.daughter[i])
	  return 1;
      return 0;
    }
  }
  

  /******************************************/
  /*                                        */
  /* Method: readTerm                       */
  /*                                        */
  /******************************************/

  private String readTerm( PushbackInputStream file ) throws IOException {
    StringBuffer buffer = new StringBuffer();
    int c;

    while (Character.isWhitespace((char)(c = file.read()))) ;
    if (c != '"')
      file.unread(c);
    else {
      for(;;) {
	c = file.read();
	if (c == '\\')
	  c = file.read();
	else if (c == '"')
	  break;
	else if (c == '\t' || c == '\n')
	  throw new IOException("Error in input file: incomplete string");
	buffer.append((char)c);
      }
    }
    return buffer.toString();
  }

  
  /******************************************/
  /*                                        */
  /* Method: readString                     */
  /*                                        */
  /******************************************/

  private String readString( PushbackInputStream file ) throws IOException {
    StringBuffer buffer = new StringBuffer();
    int c;

    /* skip preceding whitespace */
    while ((c = file.read()) != '\t' && Character.isWhitespace((char)c));

    if (c != -1 && c != '\t') {
      /* read string */
      while (c != '\t' && c != '\n' && c != '\r' && c != ' ') {
	buffer.append((char)c);
	c = file.read();
      }
      if (c == '\t')
	file.unread(c);
    }

    return buffer.toString();
  }

  /******************************************/
  /*                                        */
  /* Method: unquote                        */
  /*                                        */
  /******************************************/

  private String unquote( String s ) {
    if (s.charAt(0) == '"' && s.charAt(s.length()-1) == '"')
      return s.substring(1, s.length()-1);
    else
      return null;
  }
}
